import sys
import subprocess as sp
import tempfile as tmp

import toml


BUMP = "bumpline/cli.py"


def test_error_when_no_arguments_are_given():
    result = sp.run([BUMP], stdout=sp.PIPE, stderr=sp.PIPE)

    assert result.returncode == 2


def test_dont_change_version_when_release_isnt_given():
    with tmp.NamedTemporaryFile() as tmp_file:
        with open(tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.5"\n')

        result = sp.run([BUMP, tmp_file.name])
        version = toml.load(tmp_file.name)["project"]["version"]

        assert version == "1.5"


class TestIncreaseMajor:
    tmp_file = tmp.NamedTemporaryFile()

    def test_with_major(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2"

    def test_with_minor(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.5"\n')

        result = sp.run([BUMP, self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0"

    def test_with_micro(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.5.2"\n')

        result = sp.run([BUMP, self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0.0"

    def test_with_major_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2a"

    def test_with_major_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2b"

    def test_with_major_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2rc"

    def test_with_minor_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0a"

    def test_with_minor_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0b"

    def test_with_minor_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0rc"

    def test_with_micro_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0.0a"

    def test_with_micro_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0.0b"

    def test_with_micro_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "major"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "2.0.0rc"


class TestIncreaseMinor:
    tmp_file = tmp.NamedTemporaryFile()

    def test_with_major(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.1"

    def test_with_minor(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.5"\n')

        result = sp.run([BUMP, self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.6"

    def test_with_micro(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.5.2"\n')

        result = sp.run([BUMP, self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.6.0"

    def test_with_major_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.1a"

    def test_with_major_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.1b"

    def test_with_major_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.1rc"

    def test_with_minor_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5a"

    def test_with_minor_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5b"

    def test_with_minor_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5rc"

    def test_with_micro_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5.0a"

    def test_with_micro_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5.0b"

    def test_with_micro_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "minor"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5.0rc"


class TestIncreaseMicro:
    tmp_file = tmp.NamedTemporaryFile()

    def test_with_major(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.0.1"

    def test_with_minor(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.5"\n')

        result = sp.run([BUMP, self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5.1"

    def test_with_micro(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.5.2"\n')

        result = sp.run([BUMP, self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.5.3"

    def test_with_major_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.0.1a"

    def test_with_major_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.0.1b"

    def test_with_major_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.0.1rc"

    def test_with_minor_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.4.1a"

    def test_with_minor_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.4.1b"

    def test_with_minor_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.4.1rc"

    def test_with_micro_add_alpha(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--alpha", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.4.10a"

    def test_with_micro_add_beta(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--beta", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.4.10b"

    def test_with_micro_add_rc(self):
        with open(self.tmp_file.name, "w") as file:
            file.write('[project]\nversion = "1.4.9"\n')

        result = sp.run([BUMP, "--rc", self.tmp_file.name, "micro"])
        version = toml.load(self.tmp_file.name)["project"]["version"]

        assert version == "1.4.10rc"



import re


class Version:
    _release_expression = re.compile(r"\d+(\.\d+){0,2}")
    _pre_release_expression = re.compile(r"\D+")
    _levels = {"a": 1, "b": 2, "c": 3, "alpha": 1, "beta": 2, "rc": 3, "": 4}

    def __init__(self, version: str):
        exp = self._release_expression
        index = exp.match(version).end()

        release = version[:index]
        pre_release = version[index:]

        self._release_size = 3
        major, minor, micro, *_ = *release.split("."), None, None

        if micro is None:
            micro = "0"
            self._release_size -= 1

        if minor is None:
            minor = "0"
            self._release_size -= 1

        exp = self._pre_release_expression
        if exp.match(pre_release):
            index = exp.match(pre_release).end()

            pre_letter = pre_release[:index]
            pre_number = pre_release[index:]
        else:
            pre_letter = ""
            pre_number = ""

        self.major = int(major)
        self.minor = int(minor)
        self.micro = int(micro)
        self._pre_letter = pre_letter
        self._pre_number = int(pre_number) if pre_number != "" else 0

    def __repr__(self):
        return self.string

    def __str__(self):
        return self.string

    def __eq__(self, other):
        if self.major != other.major:
            return False
        if self.minor != other.minor:
            return False
        if self.micro != other.micro:
            return False
        if not self.pre_release_is_equal(other):
            return False

        return True

    def __lt__(self, other):
        if self.major != other.major:
            return self.major < other.major
        if self.minor != other.minor:
            return self.minor < other.minor
        if self.micro != other.micro:
            return self.micro < other.micro
        if not self.pre_release_is_equal(other):
            return self.pre_release_is_lower(other)

        return False

    def __le__(self, other):
        if self.major != other.major:
            return self.major < other.major
        if self.minor != other.minor:
            return self.minor < other.minor
        if self.micro != other.micro:
            return self.micro < other.micro
        if not self.pre_release_is_equal(other):
            return self.pre_release_is_lower(other)

        return True

    def __gt__(self, other):
        if self.major != other.major:
            return self.major > other.major
        if self.minor != other.minor:
            return self.minor > other.minor
        if self.micro != other.micro:
            return self.micro > other.micro
        if not self.pre_release_is_equal(other):
            return self.pre_release_is_greater(other)

        return False

    def __ge__(self, other):
        if self.major != other.major:
            return self.major > other.major
        if self.minor != other.minor:
            return self.minor > other.minor
        if self.micro != other.micro:
            return self.micro > other.micro
        if not self.pre_release_is_equal(other):
            return self.pre_release_is_greater(other)

        return True

    def split_pre_release(self):
        if self.pre_release == "":
            return "", 0

        exp = self._pre_release_expression
        if exp.match(self.pre_release):
            index = exp.match(self.pre_release).end()

            pre_letter = self.pre_release[:index]
            pre_number = self.pre_release[index:]
            pre_number = int(pre_number) if pre_number != "" else 0
        else:
            pre_letter = ""
            pre_number = 0

        return pre_letter, pre_number

    def pre_release_is_greater(self, other):
        this = self.split_pre_release()
        that = other.split_pre_release()
        levels = self._levels

        if levels[this[0]] != levels[that[0]]:
            return levels[this[0]] > levels[that[0]]
        if this[1] != that[1]:
            return this[1] > that[1]

        return False

    def pre_release_is_lower(self, other):
        this = self.split_pre_release()
        that = other.split_pre_release()
        levels = self._levels

        if levels[this[0]] != levels[that[0]]:
            return levels[this[0]] < levels[that[0]]
        if this[1] != that[1]:
            return this[1] < that[1]

        return False

    def pre_release_is_equal(self, other):
        this = self.split_pre_release()
        that = other.split_pre_release()
        levels = self._levels

        if levels[this[0]] != levels[that[0]]:
            return False
        if this[1] != that[1]:
            return False

        return True

    def bump_major(self):
        self.major += 1
        self.minor = 0
        self.micro = 0
        del self.pre_release

    def bump_minor(self):
        self._release_size = max(self._release_size, 2)
        self.minor += 1
        self.micro = 0
        del self.pre_release

    def bump_micro(self):
        self._release_size = max(self._release_size, 3)
        self.micro += 1
        del self.pre_release

    def add_alpha(self):
        pre_letter, pre_number = self.split_pre_release()
        if pre_letter in ["a", "alpha"]:
            pre_number += 1
        else:
            pre_letter = "a"
            pre_number = 0

        self._pre_letter = pre_letter
        self._pre_number = pre_number

    def add_beta(self):
        pre_letter, pre_number = self.split_pre_release()
        if pre_letter in ["b", "beta"]:
            pre_number += 1
        else:
            pre_letter = "b"
            pre_number = 0

        self._pre_letter = pre_letter
        self._pre_number = pre_number

    def add_rc(self):
        pre_letter, pre_number = self.split_pre_release()
        if pre_letter in ["c", "rc"]:
            pre_number += 1
        else:
            pre_letter = "rc"
            pre_number = 0

        self._pre_letter = pre_letter
        self._pre_number = pre_number

    @property
    def release(self):
        "return version release as a string"
        release_list = [self.major]
        if self.minor != 0 or self._release_size >= 2:
            release_list.append(self.minor)
        if self.micro != 0 or self._release_size == 3:
            release_list.append(self.micro)

        return ".".join(list(map(str, release_list)))

    @property
    def pre_release(self):
        "return version pre-release label"
        pre_letter = self._pre_letter
        pre_number = str(self._pre_number) if self._pre_number != 0 else ""

        return f"{pre_letter}{pre_number}"

    @pre_release.setter
    def pre_release(self, value):
        exp = self._pre_release_expression

        if exp.match(value):
            index = exp.match(value).end()

            pre_letter = value[:index]
            pre_number = value[index:]
        else:
            pre_letter = ""
            pre_number = ""

        self._pre_letter = pre_letter
        self._pre_number = int(pre_number) if pre_number != "" else 0

    @pre_release.deleter
    def pre_release(self):
        self._pre_letter = ""
        self._pre_number = 0

    @property
    def tuple(self):
        "return version in tuple format"
        return self.major, self.minor, self.micro, self.pre_release

    @tuple.setter
    def tuple(self, value):
        major, minor, micro, pre_release = list(value)

        if value[1] is None and value[2] is None:
            self._release_size = 1
            minor = 0
            micro = 0
        elif value[2] is None:
            self._release_size = 2
            micro = 0
        elif None in value:
            raise ValueError("None is only valid for micro or both minor and micro")

        self.major = major
        self.minor = minor
        self.micro = micro
        self.pre_release = pre_release

    @property
    def string(self):
        "return version in string format"
        return f"{self.release}{self.pre_release}"

    @string.setter
    def string(self, value):
        self.__init__(value)



#!/usr/bin/env python3
import os
import sys
import platform
import subprocess as sp
import pty
import argparse


def get_python_version(tty=False):
    implementation = platform.python_implementation()
    version = platform.python_version()

    if tty:
        blue = "\x1b[34m"
        yellow = "\x1b[33m"
        reset = "\x1b[00m"
        return f"{blue}{implementation} version: {yellow}{version}{reset}"
    else:
        return " ".join([implementation, "version:", version])


def write_test_to_tty():
    exit_code = 0
    # Run the command in a pty
    pid, fd = pty.fork()
    if pid == 0:
        # Child process
        status = sp.call(["make", "test"])
        os._exit(status)
    else:
        # Parent process
        try:
            while True:
                data = os.read(fd, 1024)
                if not data:
                    break
                sys.stdout.write(data.decode('utf-8'))
            _, status = os.waitpid(pid, 0)
            exit_code = os.WEXITSTATUS(status)
        except OSError:
            pass

    print(get_python_version(True))
    return exit_code


def write_test_to_file():
    command = sp.run(["make", "test"], stdout=sp.PIPE)
    print(command.stdout.decode("UTF-8").strip())
    print(get_python_version())
    return command.returncode


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--redirect-stdout", action="store_true")

    should_redirect = parser.parse_args().redirect_stdout

    if should_redirect:
        status = write_test_to_file()
    else:
        status = write_test_to_tty()

    return status


if __name__ == "__main__":
    sys.exit(main())



from bumpline.version import Version


class TestCompareVersion:
    def test_compare_equal_major_versions(self):
        assert Version("2") == Version("2.0")
        assert Version("2") == Version("2.0.0")

        assert Version("1.0a") == Version("1.0.0alpha")
        assert Version("1.2b") == Version("1.2.0beta")
        assert Version("2c") == Version("2.0rc")

    def test_compare_lower_major_versions(self):
        assert Version("1") < Version("2")
        assert Version("1") < Version("2.1")

        assert Version("1.5") < Version("2.1")
        assert Version("2.10a") < Version("3")



#!/usr/bin/env python3
import sys
import re
import argparse

import toml

from version import Version


class Release(argparse.Namespace):
    def __init__(self, args):
        self.file = args.file
        self.release = args.release
        self.pre_release = args.alpha or args.beta or args.rc


def read_file(input_file):
    content = []

    with open(input_file, "r") as file:
        for line in file.readlines():
            content.append(line)

    return content


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", help="the config file to be changed")
    parser.add_argument("release", nargs="?",
                        help="the part of the release to be increased")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-a", "--alpha", action="store_const", const="alpha",
                       help="label the new version as an alpha pre-release")
    group.add_argument("-b", "--beta", action="store_const", const="beta",
                       help="label the new version as a beta pre-release")
    group.add_argument("-c", "--rc", action="store_const", const="rc",
                       help="label the new version as a release candidate")

    args = Release(parser.parse_args())

    input_file = args.file
    content = read_file(input_file)

    get_version = lambda line: line.strip().find("version") == 0
    has_version = list(map(get_version, content))

    version_index = has_version.index(True)
    version_object = toml.loads(content[version_index])
    version = Version(version_object["version"])

    if args.release == "major":
        version.bump_major()
    elif args.release == "minor":
        version.bump_minor()
    elif args.release == "micro":
        version.bump_micro()

    if args.pre_release == "alpha":
        version.add_alpha()
    elif args.pre_release == "beta":
        version.add_beta()
    elif args.pre_release == "rc":
        version.add_rc()

    content[version_index] = toml.dumps({"version": str(version)})

    with open(input_file, "w") as file:
        for line in content:
            file.write(line)


if __name__ == "__main__":
    main()
